@@ -6,7 +6,11 @@ use warnings FATAL => 'all';
 use parent qw(Exporter);
 use Carp qw(croak);
 
-our @EXPORT_OK = qw(slist spath spath_delta);
+our @EXPORT_OK = qw(
+    slist
+    spath
+    spath_delta
+);
 
 =head1 NAME
 
@@ -14,11 +18,11 @@ Struct::Path - Path for nested structures where path is also a structure
 
 =head1 VERSION
 
-Version 0.60
+Version 0.61
 
 =cut
 
-our $VERSION = '0.60';
+our $VERSION = '0.61';
 
 =head1 SYNOPSIS
 
@@ -86,31 +90,33 @@ Don't dive into structure deeper than defined level.
 
 sub slist($;@) {
     my ($struct, %opts) = @_;
-    my @out = [[], $struct]; # init: [path, lastref]
 
-    my $continue = 1;
-    my $depth = 0;
-    while ($continue) {
-        last if (defined $opts{depth} and $depth >= $opts{depth});
-        $continue = 0;
-        my @new;
-        while (my $path = shift @out) {
-            if (ref $path->[1] eq 'ARRAY' and @{$path->[1]}) {
-                for (my $i = 0; $i < @{$path->[1]}; $i++) {
-                    push @new, [[@{$path->[0]}, [$i]], $path->[1]->[$i]];
-                }
-                $continue = 1;
-            } elsif (ref $path->[1] eq 'HASH' and keys %{$path->[1]}) {
-                for my $k (sort keys %{$path->[1]}) {
-                    push @new, [[@{$path->[0]}, {keys => [$k]}], $path->[1]->{$k}];
-                }
-                $continue = 1;
-            } else {
-                push @new, $path; # complete path
+    my @in = [[], $struct]; # init: [path, lastref]
+    return @in if (defined $opts{depth} and $opts{depth} < 1);
+
+    my @out;
+
+    while (my $p = shift @in) {
+        my @unres; # not fully resolved
+
+        if (ref $p->[1] eq 'HASH' and keys %{$p->[1]}) {
+            for my $k (sort keys %{$p->[1]}) {
+                push @unres, [[@{$p->[0]}, {keys => [$k]}], $p->[1]->{$k}];
             }
+        } elsif (ref $p->[1] eq 'ARRAY' and @{$p->[1]}) {
+            for (my $i = 0; $i < @{$p->[1]}; $i++) {
+                push @unres, [[@{$p->[0]}, [$i]], $p->[1]->[$i]];
+            }
+        } else {
+            push @out, $p;
         }
-        @out = @new;
-        $depth++;
+
+        if (defined $opts{depth} and @unres and @{$unres[0]->[0]} >= $opts{depth}) {
+            push @out, @unres;
+            next;
+        }
+
+        unshift @in, @unres; # reiterate
     }
 
     return @out;
@@ -129,7 +135,7 @@ Path is a list of 'steps', each represents nested level in structure.
 Arrayref as a step stands for ARRAY in structure and must contain desired indexes or be
 empty (means "all items"). Sequence for indexes is important and defines result sequence.
 
-Almost the same for HASHES: step must be a hashref, must contain key C<keys> which
+Almost the same for HASHes: step must be a hashref, must contain key C<keys> which
 value must contain list of desired keys in structure. Empty list means all keys. Sequence
 in C<keys> list defines result sequence.
 
@@ -140,7 +146,8 @@ expected as output.
 Why existed *Path* libs (L</"SEE ALSO">) not enough?
 This scheme has no collisions for paths like '/a/0/c' ('0' may be an ARRAY index or a key
 for HASH, depends on passed structure). In some cases this is important, for example, when
-you want to define exact path in structure, but unable to validate it's schema.
+you want to define exact path in structure, but unable to validate it's schema or when structure
+doesn't exists yet (see L</expand> for more info).
 
 See L<Struct::Path::PerlStyle> if you're looking for human friendly path definition method.
 
@@ -253,10 +260,7 @@ sub spath($$;@) {
                 }
             }
         } elsif (ref $step eq 'CODE') {
-            for my $r (@out) {
-                $step->($r->[0], $r->[1]) or croak "Failed to apply user defined function (step #$sc)";
-                push @new, $r;
-            }
+            map { $step->($_->[0], $_->[1]) and push(@new, $_) } @out;
         } else {
             croak "Unsupported thing in the path (step #$sc)";
         }
@@ -284,7 +288,9 @@ sub spath_delta($$) {
     croak "First path may be undef or an arrayref" unless (ref $frst eq 'ARRAY');
 
     my $i = 0;
-    MAIN: while ($i < @{$frst}) {
+
+    MAIN:
+    while ($i < @{$frst}) {
         last unless (ref $frst->[$i] eq ref $scnd->[$i]);
         if (ref $frst->[$i] eq 'ARRAY') {
             last unless (@{$frst->[$i]} == @{$scnd->[$i]});
@@ -360,7 +366,7 @@ L<Struct::Diff> L<Struct::Path::PerlStyle>
 
 =head1 LICENSE AND COPYRIGHT
 
-Copyright 2016 Michael Samoglyadov.
+Copyright 2016,2017 Michael Samoglyadov.
 
 This program is free software; you can redistribute it and/or modify it
 under the terms of either: the GNU General Public License as published
